#include "Plugin.h"
#include <assert.h>
#include "ffgl/ParamBase.h"
#include "ffgl/ParamBuffer.h"
#include "ffgl/ParamInteger.h"
#include "ffgl/ParamOption.h"
#include "ffgl/ParamRange.h"
#include "ParamFFT.h"
#include "ParamTrigger.h"

#define LOGSHADERNOTES

using namespace ffgl;
using namespace ffglex;

namespace ffglqs
{

Plugin::Plugin()
{
}

Plugin::~Plugin()
{
}

FFResult Plugin::InitGL( const ViewportStruct* viewPort )
{
	std::string fragmentShaderCode = CreateFragmentShader( fragmentShaderBase );
	if( !shader.Compile( vertexShaderCode, fragmentShaderCode ) )
	{
		DeinitGL();
		return FF_FAIL;
	}

	shader.Use();
	SetParamUniformLocations();

	if( !quad.Initialise() )
	{
		DeinitGL();
		return FF_FAIL;
	}

	if( Init() == FF_FAIL )
	{
		DeinitGL();
		return FF_FAIL;
	}

	//FFGL requires us to leave the context in a default state on return
	resetOpenGLState();

	//Use base-class init as success result so that it retains the viewport.
	return PluginBase::InitGL( viewPort );
}

FFResult Plugin::ProcessOpenGL( ProcessOpenGLStruct* inputTextures )
{
	UpdateAudioAndTime();
	shader.Use();
	SendDefaultParams( shader );
	SendParams( shader );
	Update();
	FFResult result = Render( inputTextures );
	ConsumeAllTriggers();
	resetOpenGLState();

	return result;
}

FFResult Plugin::DeinitGL()
{
	shader.FreeGLResources();
	quad.Release();
	Clean();
	return FF_SUCCESS;
}

FFResult Plugin::Render( ProcessOpenGLStruct* inputTextures )
{
	shader.Use();
	quad.Draw();
	return FF_SUCCESS;
}

std::string Plugin::CreateFragmentShader( const std::string& base ) const
{
	std::string fragmentShaderCode = fragmentShaderCodeStart;

	int i = 0;
	while( i < params.size() )
	{
		const ParamBase* param         = params[ i ];
		const std::string& uniformName = param->GetUniformName();

		if( uniformName.empty() )
		{
			// No sense in adding a uniform if there is no name specified!
			i++;
			continue;
		}

		const ParameterType type = param->GetType();

		if( IsRGBColor( i ) )
		{
			fragmentShaderCode += "uniform vec3 " + uniformName + ";\n";
			i += 3;
		}
		else if( IsHueColor( i ) )
		{
			fragmentShaderCode += "uniform vec4 " + uniformName + ";\n";
			i += 4;
		}
		else if( type == ParameterType::FF_TYPE_BOOLEAN || type == ParameterType::FF_TYPE_EVENT )
		{
			fragmentShaderCode += "uniform bool " + uniformName + ";\n";
			i++;
		}
		else if( type == ParameterType::FF_TYPE_INTEGER )
		{
			fragmentShaderCode += "uniform int " + uniformName + ";\n";
			i++;
		}
		else if( type == ParameterType::FF_TYPE_OPTION )
		{
			// We could handle this in the final "else" case; making explicit case to match with SendParams()
			fragmentShaderCode += "uniform float " + uniformName + ";\n";
			i++;
		}
		else if( type == ParameterType::FF_TYPE_BUFFER )
		{
			auto paramBuff = dynamic_cast< const ParamBuffer* >( param );
			if( paramBuff )
			{
				const ParameterUsage usage = paramBuff->GetUsage();

				// Check usage
				if( usage == ParameterUsage::FF_USAGE_FFT )
				{
					// Passed as a vec4 (volume, bass, mid, high)
					fragmentShaderCode += "uniform vec4 " + uniformName + ";\n";
				}
				else
				{
					// TODO: ParameterUsage::FF_USAGE_STANDARD not handled
				}
			}

			i++;
		}
		else if( type == ParameterType::FF_TYPE_TEXT )
		{
			// Text parameters not passed to shader
			// Also shouldn't have uniformName set, but just in case it is, we still want to skip it
			i++;
		}
		else
		{
			// "Normal" values (floats, 0.0 to 1.0, including non-grouped colour parameters), float range values
			fragmentShaderCode += "uniform float " + uniformName + ";\n";
			i++;
		}
	}

	for( auto snippet : includedSnippets )
	{
		fragmentShaderCode += shader::snippets.find( snippet )->second;
	}

	fragmentShaderCode += base;
	return fragmentShaderCode;
}

void Plugin::SetParamUniformLocations()
{
	for( auto param : params )
	{
		const std::string& uniformName = param->GetUniformName();
		if( !uniformName.empty() )
		{
			const GLint location = shader.FindUniform( uniformName );

			if( location != -1 )
			{
				param->SetUniformLocation( location );
			}
			else
			{
#ifdef LOGSHADERNOTES
				printf( "Note: Parameter [%s]: no shader uniform [%s]\n", param->GetName().c_str(), uniformName.c_str() );
#endif
			}
		}
	}
}

void Plugin::UpdateAudioAndTime()
{
	// Update time and frame data
	frame++;
	auto t_now = std::chrono::high_resolution_clock::now();
	timeNow    = std::chrono::duration< float, std::milli >( t_now - t_start ).count() / 1000.0f;
	deltaTime  = timeNow - lastUpdate;
	lastUpdate = timeNow;

	// Update FFT data
	for( auto paramFFT : audioParams )
	{
		paramFFT->audio.Update( paramFFT->data );
	}
}

// 2019-07-12 ffgl@subpixels.com
// - handle ParameterType::FF_TYPE_BUFFER
// - handle ParameterType::FF_TYPE_OPTION
void Plugin::SendParams( Shader& shader ) const
{
	// Clamp to edge is broken in Resolume right now so disable it
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER );
	int i = 0;
	while( i < params.size() )
	{
		const ParamBase* param      = params[ i ];
		const ParameterType type    = param->GetType();
		const GLint uniformLocation = param->GetUniformLocation();

		if( IsRGBColor( i ) )
		{
			// Uniform is taken from the first of three parameters
			if( uniformLocation != -1 )
			{
				// Using the RGB elements only; no alpha
				RGBAColor color;

				// Access the current and two following params
				color.r = ( dynamic_cast< const ParamFloat* >( GetParam( i + 0 ) ) )->GetFloatValue();
				color.g = ( dynamic_cast< const ParamFloat* >( GetParam( i + 1 ) ) )->GetFloatValue();
				color.b = ( dynamic_cast< const ParamFloat* >( GetParam( i + 2 ) ) )->GetFloatValue();

				shader.Set( uniformLocation, color.r, color.g, color.b );
			}

			i += 3;
		}
		else if( IsHueColor( i ) )
		{
			// Uniform is taken from the first of four parameters
			if( uniformLocation != -1 )
			{
				HSBAColor color;

				// Access the current and three following params
				color.hue   = ( dynamic_cast< const ParamFloat* >( GetParam( i + 0 ) ) )->GetFloatValue();
				color.sat   = ( dynamic_cast< const ParamFloat* >( GetParam( i + 1 ) ) )->GetFloatValue();
				color.bri   = ( dynamic_cast< const ParamFloat* >( GetParam( i + 2 ) ) )->GetFloatValue();
				color.alpha = ( dynamic_cast< const ParamFloat* >( GetParam( i + 3 ) ) )->GetFloatValue();

				//we need to make sure the hue doesn't reach 1.0f, otherwise the result will be pink and not red how it should be
				if( color.hue == 1.0f )
					color.hue = 0.0f;

				RGBAColor rgba;
				HSBtoRGB( color, rgba );

				shader.Set( uniformLocation, rgba.r, rgba.g, rgba.b, rgba.a );
			}

			i += 4;
		}
		else if( type == ParameterType::FF_TYPE_BOOLEAN || type == ParameterType::FF_TYPE_EVENT )
		{
			if( uniformLocation != -1 )
			{
				auto paramBool = dynamic_cast< const ParamBool* >( param );
				if( paramBool )
					shader.Set( uniformLocation, static_cast< int >( paramBool->GetBoolValue() ) );
			}

			i++;
		}
		else if( type == ParameterType::FF_TYPE_INTEGER )
		{
			// Integers are defined with a range, but the value is not normalised (0.0 to 1.0)
			if( uniformLocation != -1 )
			{
				auto paramInteger = dynamic_cast< const ParamInteger* >( param );
				if( paramInteger )
					shader.Set( uniformLocation, paramInteger->GetIntegerValue() );
			}

			i++;
		}
		else if( type == ParameterType::FF_TYPE_OPTION )
		{
			if( uniformLocation != -1 )
			{
				auto paramOption = dynamic_cast< const ParamOption* >( param );
				if( paramOption )
					shader.Set( uniformLocation, paramOption->GetFloatValue() );
			}

			i++;
		}
		else if( type == ParameterType::FF_TYPE_BUFFER )
		{
			if( uniformLocation != -1 )
			{
				auto paramBuffer = dynamic_cast< const ParamBuffer* >( param );
				if( paramBuffer )
				{
					// Check usage
					if( paramBuffer->GetUsage() == ParameterUsage::FF_USAGE_FFT )
					{
						// Note: Buffer data (elements[]) is copied to fftData[] in UpdateAudioAndTime()
						const auto paramFFT = dynamic_cast< const ParamFFT* >( param );
						if( paramFFT )
						{
							const Audio& a = paramFFT->audio;
							// Passed as a vec4 (volume, bass, mid, high)
							shader.Set( uniformLocation, a.GetVolume(), a.GetBass(), a.GetMid(), a.GetHigh() );
						}
					}
					else
					{
						// TODO: ParameterUsage::FF_USAGE_STANDARD not handled
					}
				}
			}

			i++;
		}
		else if( type == ParameterType::FF_TYPE_TEXT )
		{
			// Text parameters not passed to shader
			i++;
		}
		else
		{
			// "Normal" values (floats, 0.0 to 1.0, including non-grouped colour parameters), float range values
			if( uniformLocation != -1 )
			{
				//auto paramFloat = dynamic_cast< const ParamFloat* >( param );
				//if( paramFloat )
				//{
				//	auto paramRange = dynamic_cast< const ParamRange* >( param );
				//	float value     = paramRange ? paramRange->GetRealValue() : paramFloat->GetFloatValue();
				//	shader.Set( uniformLocation, value );
				//}

				auto paramFloat = dynamic_cast< const ParamFloat* >( param );
				if( paramFloat )
					shader.Set( uniformLocation, paramFloat->GetFloatValue() );
			}

			i++;
		}
	}
}

void Plugin::SendDefaultParams( ffglex::Shader& shader ) const
{
	shader.Set( "time", timeNow );
	shader.Set( "deltaTime", deltaTime );
	shader.Set( "frame", frame );
	shader.Set( "resolution", (float)viewport.width, (float)viewport.height );
	shader.Set( "bpm", beatInfo.bpm );
	shader.Set( "phase", beatInfo.barPhase );
}

//FFResult Plugin::SetFloatParameter( unsigned int index, float value )
//{
//	if( index < params.size() )
//	{
//		params[ index ]->setValue( value );
//		return FF_SUCCESS;
//	}
//	else
//	{
//		return FF_FAIL;
//	}
//}

//float Plugin::GetFloatParameter( unsigned int index )
//{
//	if( 0 <= index && index < params.size() )
//	{
//		return params[ index ]->getValue();
//	}
//	else
//	{
//		return 0.0f;
//	}
//}

//// 2019-07-17 ffgl@subpixels.com - use setText()
//FFResult Plugin::SetTextParameter( unsigned int index, const char* value )
//{
//	bool inBounds = 0 <= index && index < params.size();
//	if( !inBounds )
//		return FF_FAIL;
//
//	auto paramText = dynamic_cast< ParamText* >( params[ index ] );
//	if( !paramText )
//		return FF_FAIL;
//
//	paramText->setText( value == nullptr ? "" : value );
//	return FF_SUCCESS;
//}

//// 2019-07-17 ffgl@subpixels.com - use getText()
//char* Plugin::GetTextParameter( unsigned int index )
//{
//	bool inBounds = 0 <= index && index < params.size();
//	if( !inBounds )
//		return "";
//
//	auto paramText = dynamic_cast< ParamText* >( params[ index ] );
//	if( !paramText )
//		return "";
//
//	return const_cast< char* >( paramText->getText().c_str() );
//}

void Plugin::SetSampleRate( unsigned int _sampleRate )
{
	sampleRate = _sampleRate;
	for( auto paramFFT : audioParams )
	{
		paramFFT->audio.SetSampleRate( _sampleRate );
	}
}

void Plugin::SetFragmentShader( const std::string& base )
{
	fragmentShaderBase = base;
}

void Plugin::AddAudioParam( ParamFFT* param )
{
	PluginManager::AddParam( param );
	audioParams.push_back( param );
}

ParamHSBAStruct Plugin::AddHueColorParam( const std::string& name, const HSBAColor& defaultColor )
{
	return AddHueColorParam( name, name, defaultColor );
}

ParamHSBAStruct Plugin::AddHueColorParam( const std::string& name, const std::string& uniformName, const HSBAColor& defaultColor )
{
	ParamHSBAStruct hsba;

	hsba.hue   = new ParamFloat( name, ParameterType::FF_TYPE_HUE, uniformName, defaultColor.hue );
	hsba.sat   = new ParamFloat( name + " S", ParameterType::FF_TYPE_SATURATION, {}, defaultColor.sat );
	hsba.bri   = new ParamFloat( name + " B", ParameterType::FF_TYPE_BRIGHTNESS, {}, defaultColor.bri );
	hsba.alpha = new ParamFloat( name + " A", ParameterType::FF_TYPE_ALPHA, {}, defaultColor.alpha );

	AddParam( hsba.hue );
	AddParam( hsba.sat );
	AddParam( hsba.bri );
	AddParam( hsba.alpha );

	return hsba;
}

// 2019-08-13 ffgl@subpixels.com - Body extracted to AddRGBColorParam(name, uniformName) constructor
ParamRGBStruct Plugin::AddRGBColorParam( const std::string& name, const ffglex::RGBAColor& defaultColor )
{
	return AddRGBColorParam( name, name, defaultColor );
}

// 2019-08-13 ffgl@subpixels.com
// - Change plugin paramater titles from ("name", "name_g", "name_b") to ("name R", "name G", "name B")
// - Use Param( name, type, uniformName, defaultValue ) constructor, no uniform names for the G and B parameters
ParamRGBStruct Plugin::AddRGBColorParam( const std::string& name, const std::string& uniformName, const ffglex::RGBAColor& defaultColor )
{
	ParamRGBStruct rgb;

	rgb.red   = new ParamFloat( name + " R", ParameterType::FF_TYPE_RED, uniformName, 0.5f );
	rgb.green = new ParamFloat( name + " G", ParameterType::FF_TYPE_GREEN, {}, 0.5f );
	rgb.blue  = new ParamFloat( name + " B", ParameterType::FF_TYPE_BLUE, {}, 0.5f );

	AddParam( rgb.red );
	AddParam( rgb.green );
	AddParam( rgb.blue );

	return rgb;
}

ParamRGBStruct Plugin::AddRGBRangeParam( const std::string& name, const std::string& uniformName, float defaultValue, const RangeStruct range )
{
	ParamRGBStruct rgb;

	rgb.red   = new ParamRange( name + " R", ParameterType::FF_TYPE_RED, uniformName, defaultValue, range );
	rgb.green = new ParamRange( name + " G", ParameterType::FF_TYPE_GREEN, {}, defaultValue, range );
	rgb.blue  = new ParamRange( name + " B", ParameterType::FF_TYPE_BLUE, {}, defaultValue, range );

	AddParam( rgb.red );
	AddParam( rgb.green );
	AddParam( rgb.blue );

	return rgb;
}

bool Plugin::IsHueColor( int index ) const
{
	bool enoughSpace = index + 3 < params.size();
	if( !enoughSpace )
		return false;
	bool isColorType =
		params[ index ]->GetType() == ParameterType::FF_TYPE_HUE &&
		params[ index + 1 ]->GetType() == ParameterType::FF_TYPE_SATURATION &&
		params[ index + 2 ]->GetType() == ParameterType::FF_TYPE_BRIGHTNESS &&
		params[ index + 3 ]->GetType() == ParameterType::FF_TYPE_ALPHA;

	return isColorType;
}

bool Plugin::IsRGBColor( int index ) const
{
	bool enoughSpace = index + 2 < params.size();
	if( !enoughSpace )
		return false;
	bool isColorType =
		params[ index ]->GetType() == ParameterType::FF_TYPE_RED &&
		params[ index + 1 ]->GetType() == ParameterType::FF_TYPE_GREEN &&
		params[ index + 2 ]->GetType() == ParameterType::FF_TYPE_BLUE;

	return isColorType;
}

void Plugin::Include( shader::snippet_id snippet )
{
	if( includedSnippets.find( snippet ) != includedSnippets.end() )
		return;

	includedSnippets.insert( snippet );
	auto deps = shader::dependencies.find( snippet );
	if( deps == shader::dependencies.end() )
		return;
	for( auto dep : deps->second )
	{
		includedSnippets.insert( dep );
	}
}

void Plugin::Include( std::set< shader::snippet_id > snippets )
{
	for( auto snippet : snippets )
	{
		Include( snippet );
	}
}

void Plugin::ConsumeAllTriggers()
{
	for( auto param : params )
	{
		auto trigger = dynamic_cast< ParamTrigger* >( param );
		if( trigger )
		{
			trigger->Consume();
		}
	}
}

}// namespace ffglqs
